import os.path
import zipfile
import json
from abc import ABC
from io import BytesIO
from typing import List, NamedTuple, Dict, IO

from koodous.exceptions import FeedPackageException
from koodous.utils import request_download


class Sample(NamedTuple):
    """Sample information."""
    sha256: str
    download_url: str


class Analysis(NamedTuple):
    """Analysis information."""
    sha256: str
    analysis: dict


class Feed(ABC):
    """Feed abstract class.
    """

    def __init__(self, bytes_: BytesIO, package: str):
        """

        :param bytes_: bytes with the file response content.
        :param package: package api parameter.
        """
        self.bytes = bytes_
        self.package = package

    def _unzip(self) -> Dict[str, IO[bytes]]:
        zip_file = zipfile.ZipFile(self.bytes)
        files = {}
        for file_name in zip_file.namelist():
            files[file_name] = zip_file.open(file_name)
        return files

    def write(self, output_file: str) -> None:
        """Write the feed zip to the desired file path.

        :param output_file: path to the file output.
        :return: None.
        """
        with open(output_file, 'wb') as f:
            f.write(self.bytes.getbuffer().tobytes())


class ApksFeed(Feed):
    """Class to interact with the output of the apks feed api.

    Things that can be done with this class:
        - Get the samples sha256s and their download links.
        - Download the samples apks to a folder.
    """

    def samples(self) -> List[Sample]:
        """Get the samples from the feed.

        This methods extract the zip and construct the tuples with the samples information (sha256, download url).

        :return: samples sha256s and the download url.
        """
        unzipped_file = self._unzip()
        # Check that the file samples in in the zip.
        if 'samples' not in unzipped_file:
            raise FeedPackageException('The expected file is not in the package feed zip.')
        # Extract the samples file and create the tuples with the samples.
        samples_file = unzipped_file['samples']
        samples_bytes = samples_file.readlines()
        samples = [sample_bytes.decode().strip().split(';') for sample_bytes in samples_bytes]
        return [Sample(sample[0], sample[1]) for sample in samples]

    @staticmethod
    def download_sample(sample: Sample, output_folder: str) -> None:
        """Download the apk from the sample.

        :param sample: sample (sha256, download link).
        :param output_folder: folder to save the apk.
        :return: None.
        """
        file_path = os.path.join(output_folder, f'{sample.sha256}.apk')
        # Download and store the sample.
        request_download(sample.download_url, file_path)

    def download_samples(self, samples: List[Sample], output_folder: str) -> None:
        """Download the apks from the samples

        :param samples: list of samples.
        :param output_folder: folder to save the apk.
        :return: None.
        """
        # Check that the directory exists.
        if not os.path.isdir(output_folder):
            raise FileNotFoundError(f'The directory {output_folder} does not exist.')
        # Download the samples.
        for sample in samples:
            self.download_sample(sample, output_folder)


class AnalysesFeed(Feed):
    """Class to interact with the output of the analyses feed.

    Things that can be done with this class:
    - Get the analysis reports directly in json format with ``analysis()``.
    """

    def analyses(self) -> List[Analysis]:
        """Obtain the analyses from the feed and create a list with them.

        :return: tuples with the sha256 of the analyzed sample and the analysis report of it.
        """
        unzipped_file = self._unzip()
        analyses = []
        for name, file in unzipped_file.items():
            sha256 = name[:-len('.json')]
            analysis = json.load(file)
            analyses.append(Analysis(sha256, analysis))
        return analyses



from koodous.representations.base import BaseRepresentation


class Rule(BaseRepresentation):
    """The rule representation.
    """

    def __init__(self):
        raise NotImplementedError()  # TODO



class ApiException(Exception):
    """Koodous API base class."""


class ApiUnauthorizedException(ApiException):
    """Exception when is made a request without the correct token or insufficient privileges."""


class FeedException(ApiException):
    """Base class for the feed exceptions."""


class FeedPackageException(FeedException):
    """Exception related to the feed package."""


class PackageInvalidFormatException(FeedException):
    """Package format invalid exception. Check the koodous api docs."""


class ApkNotFound(ApiException):
    """Apk sample not found."""



from datetime import datetime
from typing import List, Optional

from koodous.representations.base import BaseRepresentation


class Comment(BaseRepresentation):
    """The ruleset representation.
    """

    def __str__(self):
        return f'Comment(id={self.id})'

    def __repr__(self):
        return f'RuleSet()'

    @property
    def id(self) -> int:
        """Comment identifier."""
        return self.data['id']

    @property
    def author_data(self) -> dict:
        """The analyst dict data."""
        return self.data['author']

    @property
    def author(self) -> None:
        """TODO: Return the Author representation."""
        raise NotImplementedError

    @property
    def created_on(self) -> datetime:
        """The date when comment was created."""
        return self._fromtimestamp(self.data['created_on'])

    @property
    def modified_on(self) -> datetime:
        """The date when comment was modified."""
        return self._fromtimestamp(self.data['modified_on'])

    @property
    def apk(self) -> str:
        """APK sha256 hash."""
        return self.data['apk']

    @property
    def text(self) -> None:
        """The comment's content."""
        return self.data['text']



import hashlib
from datetime import datetime
from typing import Optional, Union, Dict, List
from urllib.parse import urlparse, parse_qs

import requests

from koodous.exceptions import PackageInvalidFormatException


def package_param(package: Optional[Union[str, datetime]]) -> str:
    """Validate and create the package parameter.

    Using the datetime object you can't specify a hourly package because the datetime minute is 0 when it is not
    set.

    :param package: package parameter.
    :return: package string to the request.
    """
    if isinstance(package, datetime):
        return package.strftime('%Y%m%dT%H%M')
    elif isinstance(package, str):
        return package  # TODO: Check that the string is valid.
    else:
        raise PackageInvalidFormatException('The package type is not valid. See the api docs.')


def sha256sum(file_path: str) -> str:
    """Calculate the file sha256.

    :param file_path: path to the file.
    :return: file checksum (sha256).
    """
    h = hashlib.sha256()
    b = bytearray(128*1024)
    mv = memoryview(b)
    with open(file_path, 'rb', buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()


def extract_url_parameters(url: str) -> Dict[str, List[str]]:
    """Extract the query parameters from a url string.

    :param url: the url string.
    :return: the dictionary with the query parameters.
    """
    url_parsed = urlparse(url)
    qs_parsed = parse_qs(url_parsed.query)
    return qs_parsed


def request_download(download_url: str, output_filepath: str, chunk_size: int = 8192, **kwargs) -> None:
    """Request to download a file from a url.

    This method use a simple ``GET`` request without any pre-defined headers or parameters.

    :param download_url: download url.
    :param output_filepath: full path to the write the file.
    :param chunk_size: chunk size.
    :param kwargs: kwargs to be included in the GET request.
    :return: None.
    """
    with requests.get(download_url, stream=True, **kwargs) as r:
        r.raise_for_status()
        with open(output_filepath, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)



KOODOUS_API_URL = 'https://api.koodous.com'
KOODOUS_API_ENDPOINTS = {
    # Most used methods.
    'apks': '/apks',
    'apks_detail': '/apks/{sha256}',
    'apks_detail_download': '/apks/{sha256}/download',
    'apks_detail_upload': '/apks/{sha256}/get_upload_url',
    'apks_detail_analysis': '/apks/{sha256}/analysis',
    'apks_detail_analyze': '/apks/{sha256}/analyze',
    'apks_feed': '/feed/apks?package={package}',
    'analyses_feed': '/feed/analyses?package={package}',
    'detected_feed': '/feed/detected?package={package}',
    # Misc methods.
    'apks_detail_comments': '/apks/{sha256}/comments',
    'apks_detail_comments_detail': '/apks/{sha256}/comments/{comment_id}',
    'apks_detail_votes': '/apks/{sha256}/votes',
    'analysts': '/analysts',
    'analysts_detail': '/analysts/{username}',
    'analysts_detail_followers': '/analysts/{username}/followers',
    'analysts_detail_following': '/analysts/{username}/following',
    'analysts_detail_comments': '/analysts/{username}/comments',
    'analysts_detail_activity': '/analysts/{username}/activity',
    'analysts_detail_rulesets': '/analysts/{username}/rulesets',
    'analysts_detail_follow': '/analysts/{username}/follow',
    'analysts_detail_unfollow': '/analysts/{username}/unfollow',
    'user_rulesets': '/user_rulesets',
    'user_rulesets_detail': '/user_rulesets/{rule_id}',
    'public_rulesets': '/public_rulesets',
    'public_rulesets_detail': '/public_rulesets/{ruleset_id}',
    'public_rulesets_detail_votes': '/public_rulesets/{ruleset_id}/votes',
}



from datetime import datetime
from typing import List, Optional

from koodous.representations.base import BaseRepresentation
from koodous.representations.rule import Rule


class Ruleset(BaseRepresentation):
    """The ruleset representation.
    """

    def __str__(self):
        return f'RuleSet(id={self.id}, name={self.name}, active={self.active}, deleted={self.deleted})'

    def __repr__(self):
        return f'RuleSet(id={self.id}, name={self.name})'

    @property
    def id(self) -> int:
        """The ruleset identifier."""
        return self.data['id']

    @property
    def created_on(self) -> datetime:
        """The date when the ruleset was created."""
        return self._fromtimestamp(self.data['created_on'])

    @property
    def modified_on(self) -> datetime:
        """The last date when the ruleset was modified."""
        return self._fromtimestamp(self.data['modified_on'])

    @property
    def name(self) -> str:
        """The ruleset name."""
        return self.data['name']

    @property
    def rules(self) -> List[Rule]:
        """The set of rules."""
        raise NotImplementedError  # TODO

    @property
    def rules_raw(self) -> str:
        """The set of rules in raw string."""
        return self.data['rules']

    @property
    def active(self) -> bool:
        """If the ruleset is active or not."""
        return self.data['active']

    @property
    def privacy(self) -> str:
        """The ruleset privacy parameter. Can be `public` or `private`."""
        return self.data['privacy']

    @property
    def social(self) -> bool:
        """If the ruleset is inside the Koodous system."""
        return self.data['social']

    @property
    def pending_social(self) -> bool:
        """TODO"""
        return self.data['pending_social']

    @property
    def deleted(self) -> bool:
        """If the ruleset is deleted."""
        return self.data['deleted']

    @property
    def send_notifications(self) -> bool:
        """If the ruleset will notify or not."""
        return self.data['send_notifications']

    @property
    def detections(self) -> int:
        """The number of matches done with this ruleset."""
        return self.data['detections']

    @property
    def rating(self) -> int:
        """The ruleset rating."""
        return self.data['rating']

    @property
    def parent(self) -> Optional[str]:
        """TODO"""
        return self.data['parent']



from datetime import datetime
from typing import List

from koodous.representations.base import BaseRepresentation


class Apk(BaseRepresentation):
    """Apk sample representation.
    """

    def __str__(self):
        return f'ApkSample({self.sha256}, app={self.app}, detected={self.detected})'

    def __repr__(self):
        return f'ApkSample({self.sha256}'

    @property
    def created_on(self) -> datetime:
        """The date when the apk was created on koodous."""
        return self._fromtimestamp(self.data['created_on'])

    @property
    def rating(self) -> int:
        """The value of users votes."""
        return self.data['rating']

    @property
    def image(self) -> str:
        """The app image shown on devices."""
        return self.data['image']

    @property
    def tags(self) -> List[str]:
        """The app tag list."""
        return self.data['tags']

    @property
    def md5(self) -> str:
        """The apk md5 hash."""
        return self.data['md5']

    @property
    def sha1(self) -> str:
        """The apk sha1 hash."""
        return self.data['sha1']

    @property
    def sha256(self) -> str:
        """The apk sha256 hash."""
        return self.data['sha256']

    @property
    def app(self) -> str:
        """The app name shown on devices."""
        return self.data['app']

    @property
    def package_name(self) -> str:
        """The app package identifier."""
        return self.data['package_name']

    @property
    def company(self) -> str:
        """The company or developer identifier."""
        return self.data['company']

    @property
    def developer(self) -> str:
        """The company or developer identifier."""
        return self.company

    @property
    def displayed_version(self) -> str:
        """The app version shown on market."""
        return self.data['displayed_version']

    @property
    def size(self) -> int:
        """The app size in bytes."""
        return self.data['size']

    @property
    def stored(self) -> bool:
        """TODO"""
        return self.data['stored']

    @property
    def analyzed(self) -> bool:
        """The apks has been analyzed or not by Koodous system."""
        return self.data['analyzed']

    @property
    def is_apk(self) -> bool:
        """TODO"""
        return self.data['is_apk']

    @property
    def detected(self) -> bool:
        """The apk has been detected by Koodous system."""
        return self.data['detected']

    @property
    def corrupted(self) -> bool:
        """The app cannot be executed properly."""
        return self.data['corrupted']

    @property
    def repo(self) -> str:
        """TODO"""
        return self.data['repo']

    @property
    def on_devices(self) -> bool:
        """TODO"""
        return self.data['on_devices']



"""Koodous API class."""
import os
from http import HTTPStatus
from io import BytesIO
from itertools import count
from typing import Optional, Union, Iterator, Dict, Any
from datetime import datetime

import requests

from koodous.constants import KOODOUS_API_URL, KOODOUS_API_ENDPOINTS
from koodous.exceptions import ApiUnauthorizedException, FeedPackageException, ApkNotFound
from koodous.feed import ApksFeed, AnalysesFeed
from koodous.representations.apk import Apk
from koodous.representations.comment import Comment
from koodous.utils import package_param, sha256sum, extract_url_parameters, request_download


class ApiBase:
    """Main class to interact with the different Koodous modules apis.
    """
    api_url: str
    api_endpoints: Dict[str, str]

    def request(self, method: str, endpoint: str, **kwargs) -> requests.Response:
        """Internal method to do the requests and handle possible exceptions.

        :param method: HTTP method : ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        :param endpoint: endpoint of the Koodous API.
        :param kwargs: dictionary with the arguments to the request.
        :return: the response object.
        """
        r = requests.request(method, self.api_url + endpoint, **kwargs)
        return r

    def _iterator_cursor_pagination(self, endpoint: str, params: Optional[dict] = None, **kwargs) -> Iterator[Any]:
        """Iterator over endpoints with cursor pagination.

        :param endpoint: endpoint.
        :param params: request query parameters.
        :param kwargs: dictionary with the arguments to the request.
        :return: iterator over the results data.
        """
        # Check if the params is None.
        params = params or {}
        # Iterate over the cursor pages.
        next_cursor = ''
        while next_cursor is not None:
            r_params = {
                'cursor': next_cursor,
                **params
            }
            r = self.request('GET', endpoint, params=r_params, **kwargs)
            # Check the response status code.
            r.raise_for_status()
            # Get the response and yield every apk, when it finish, go to the next cursor (page).
            response_json = r.json()
            yield from response_json['results']
            # Extract the next cursor from the fully url.
            next_cursor_url = response_json['next']
            if next_cursor_url:
                next_cursor = extract_url_parameters(next_cursor_url)['cursor'][0]
            else:
                next_cursor = None

    def _iterator_page_pagination(self, endpoint: str, params: Optional[dict] = None, **kwargs):
        """Iterator over endpoints with page pagination.

        :param endpoint: endpoint.
        :param params: request query parameters.
        :param kwargs: dictionary with the arguments to the request.
        :return: iterator over the results data.
        """
        # Check if the params is None.
        params = params or {}
        # Iterate over the pages.
        for page in count(1):
            r_params = {
                'page': page,
                **params
            }
            r = self.request('GET', endpoint, params=r_params, **kwargs)
            # Check the response status code.
            r.raise_for_status()
            # Get the response and yield every apk, when it finish, go to the next cursor (page).
            response_json = r.json()
            yield from response_json['results']
            # Return None when it's the last page.
            if response_json['next'] is None:
                return


class ApiWithTokenBase(ApiBase):
    """Proxy class to use the apis that needs token authentication.
    """

    def __init__(self, token: str, *args, **kwargs):
        """

        :param token: authentication token.
        """
        super(ApiWithTokenBase, self).__init__(*args, **kwargs)
        self.token = token

    def request(self, method: str, endpoint: str, **kwargs) -> requests.Response:
        """Internal method to do the requests and handle possible exceptions.

        :param method: HTTP method : ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        :param endpoint: endpoint of the Koodous API.
        :param kwargs: dictionary with the arguments to the request.
        :return: the response object.
        """
        # Update or add the headers to the request kwargs.
        headers = kwargs.get('headers', {})
        headers.update({'Authorization': f'Token {self.token}'})
        kwargs.update({'headers': headers})
        # Do the request and handle exceptions.
        r = super(ApiWithTokenBase, self).request(method, endpoint, **kwargs)
        if r.status_code == HTTPStatus.UNAUTHORIZED:
            raise ApiUnauthorizedException('The token is invalid or insufficient permissions.')
        return r


class Koodous(ApiWithTokenBase):
    """Main class to interact with the Koodous API.
    """
    api_url = KOODOUS_API_URL
    api_endpoints = KOODOUS_API_ENDPOINTS

    def _download_feed(self, feed: str, package: Optional[Union[str, datetime]] = '') -> requests.Response:
        """Internal method to do the response to the feed endpoint. Can choose between apks, analyses and detections.

        :param package: the package datetime parameter.
        :return: the response instance.
        """
        r = self.request('GET', feed, stream=True, params={'package': package_param(package)})
        # Check the package is available.
        if r.status_code == HTTPStatus.NOT_FOUND:
            raise FeedPackageException('Make sure the date on the package is not older than a week.')
        return r

    def apks(self, search: str = '', **params) -> Iterator[Apk]:
        """Create an apks iterator. This method is alternative to ``list_apks``.

        You can use this method iterating over the it:

        .. code-block:: python

           koodous = Koodous('KOODOUS_API_TOKEN')
           apks = koodous.apks()
           for i in range(10):
               print(next(apks))

        **Be careful with the amount of apks retrieved if you have a small request limit.**

        :param search: the parameters search.
        :param params: other parameters to be included in the request.
        :return: apks iterator.
        """
        r_params = {'search': search, **params}
        yield from (Apk(data) for data in self._iterator_cursor_pagination(self.api_endpoints['apks'], r_params))

    def apk(self, sha256: str) -> Apk:
        """Get the apk sample data.

        :param sha256: sha256 checksum.
        :return: the apk sample representation.
        """
        r = self.request('GET', self.api_endpoints['apks_detail'].format(sha256=sha256))
        # Check the response status code.
        if r.status_code == HTTPStatus.NOT_FOUND:
            raise ApkNotFound(f'The apk {sha256} not found.')
        else:
            r.raise_for_status()
        # Create the apk representation object.
        json_ = r.json()
        return Apk(json_)

    def download_apk(self, sha256: str, output_folder: str) -> str:
        """Download the apk sample.

        :param sha256: sha256 checksum.
        :param output_folder: the folder to save the apk.
        :return: the output file path.
        """
        filename = sha256 + '.apk'
        output_path = os.path.join(output_folder, filename)
        r = self.request('GET', self.api_endpoints['apks_detail_download'].format(sha256=sha256))
        # Check the response status code.
        if r.status_code == HTTPStatus.NOT_FOUND:
            raise ApkNotFound(f'The apk {sha256} not found.')
        elif r.status_code == HTTPStatus.NO_CONTENT:
            raise ApkNotFound(f'The apk {sha256} has no download link.')
        else:
            r.raise_for_status()
        # Get the download url.
        json_ = r.json()
        download_url = json_['download_url']
        # Download and store the sample.
        request_download(download_url, output_path)
        return output_path

    def upload_apk(self, file_path: str) -> None:
        """Upload an apk to the Koodous system.

        :param file_path: the apk file path.
        :return: None.
        """
        # Get the apk sha256.
        sha256 = sha256sum(file_path)
        # Get the upload url.
        r = self.request('GET', self.api_endpoints['apks_detail_upload'].format(sha256=sha256))
        # Check the response status code.
        r.raise_for_status()
        # Get the upload url.
        json_ = r.json()
        upload_url = json_['upload_url']
        # Upload the file to the given url.
        with open(file_path, 'rb') as f:
            files = {'file': f}
            r = self.request('POST', upload_url, files=files)
            r.raise_for_status()

    def analysis_apk(self, sha256: str) -> dict:
        """Get the apk analysis from Koodous system.

        :param sha256: sha256 checksum.
        :return: the analysis representation object.
        """
        r = self.request('GET', self.api_endpoints['apks_detail_analysis'].format(sha256=sha256))
        # Check the response status code.
        r.raise_for_status()
        return r.json()

    def analyze_apk(self, sha256: str) -> None:
        """Send an apk to be analyzed by the Koodous system.

        :param sha256: sha256 checksum.
        :return: None.
        """
        r = self.request('GET', self.api_endpoints['apks_detail_analyze'].format(sha256=sha256))
        # Check the response status code.
        r.raise_for_status()

    def feed_apks(self, package: Optional[Union[str, datetime]] = '') -> ApksFeed:
        """Get the apks feed.

        Note that using the datetime object you can't specify a hourly package because the datetime minute is 0 when it
        is not set.

        :param package: the package datetime. Can be a str with the format %Y%m%dT%H%M or %Y%m%dT%H or a datetime
        object.
        :return: the apks feed instance.
        """
        r = self._download_feed(self.api_endpoints['apks_feed'], package)
        return ApksFeed(BytesIO(r.raw.read(decode_content=True)), package)

    def feed_analyses(self, package: Optional[Union[str, datetime]] = '') -> AnalysesFeed:
        """Get the analyses feed.

        Note that using the datetime object you can't specify a hourly package because the datetime minute is 0 when it
        is not set.

        :param package: the package datetime. Can be a str with the format %Y%m%dT%H%M or %Y%m%dT%H or a datetime
        object.
        :return: the apk analyses instance.
        """
        r = self._download_feed(self.api_endpoints['analyses_feed'], package)
        return AnalysesFeed(BytesIO(r.raw.read(decode_content=True)), package)

    def feed_detected(self, package: Optional[Union[str, datetime]] = '') -> AnalysesFeed:
        """Get the analysis feed from the detected samples only.

        Note that using the datetime object you can't specify a hourly package because the datetime minute is 0 when it
        is not set.

        :param package: the package datetime. Can be a str with the format %Y%m%dT%H%M or %Y%m%dT%H or a datetime
        object.
        :return: the analysis feed instance.
        """
        r = self._download_feed(self.api_endpoints['detected_feed'], package)
        return AnalysesFeed(BytesIO(r.raw.read(decode_content=True)), package)

    def apk_comments(self, sha256: str) -> Iterator[Comment]:
        """Get the comments from an apk.

        :param sha256: apk checksum.
        :return: the apk comments.
        """
        yield from self._iterator_cursor_pagination(self.api_endpoints['apks_detail_comments'].format(sha256=sha256))

    def apk_create_comment(self, sha256: str, text: str) -> None:
        """Create a comment in an apk.

        :param sha256: apk checksum.
        :param text: the comment text.
        :return: None.
        """
        r = self.request('POST', self.api_endpoints['apks_detail_comments'].format(sha256=sha256), data={'text': text})
        r.raise_for_status()

    def apk_update_comment(self, sha256: str, comment_id: int, text: str) -> None:
        """Update a comment in an apk (you must be the owner or have enough permissions).

        :param sha256: apk checksum.
        :param comment_id: comment identifier.
        :param text: the new comment text.
        :return: None.
        """
        endpoint = self.api_endpoints['apks_detail_comments_detail'].format(sha256=sha256, comment_id=comment_id)
        r = self.request('PUT', endpoint, data={'text': text})
        r.raise_for_status()

    def apk_delete_comment(self, sha256: str, comment_id: str) -> None:
        """Delete a comment in a apk (you must be the owner or have enough permissions).

        :param sha256: apk checksum.
        :param comment_id: comment identifier.
        :return: None.
        """
        endpoint = self.api_endpoints['apks_detail_comments_detail'].format(sha256=sha256, comment_id=comment_id)
        r = self.request('DELETE', endpoint)
        r.raise_for_status()



from datetime import datetime, timezone


class BaseRepresentation:
    """API objects base representation."""

    def __init__(self, data: dict):
        """

        :param data the representation data.
        """
        self.data = data

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.data == self.data
        return False

    def __hash__(self):
        return hash(self.data.values())

    @staticmethod
    def _fromtimestamp(timestamp):
        return datetime.fromtimestamp(timestamp, tz=timezone.utc)
